import time 
import torch
torch.set_num_threads(4)
from lanczos import lanczos


def mpsrg(A, T, lanczos_steps=0):

    Asymm = (A + A.permute(2, 1, 0))*0.5
    D, d = Asymm.shape[0], Asymm.shape[1]
    #t0 = time.time()
    
    if lanczos_steps>0:
        #phi0 = Asymm.view(D**2*d)
        phi0 = torch.randn(D**2*d, dtype=T.dtype, device=T.device)
        phi0 = phi0/phi0.norm()
        def Hopt(x):
            Tx = (T.view(-1, d) @ x.view(D, d, D).permute(1, 0, 2).contiguous().view(d,-1)).view(d,d,d,D,D).permute(1,3,0,2,4).contiguous()
            return ((Asymm.view(D, d*D)@Tx.view(d*D, d*d*D)).view(D*d, d*D)@Asymm.permute(1,2,0).contiguous().view(d*D, D)).view(D**2*d)

        w = lanczos(Hopt, phi0, lanczos_steps)
    else:
        B = torch.einsum('ldr,adcb,icj->lairbj', (Asymm, T, Asymm)).contiguous().view(D**2*d, D**2*d)
        w, _ = torch.symeig(B, eigenvectors=True) 
    lnZ1 = torch.log(w.abs().max())
    
    if lanczos_steps>0:
        #phi0 = Asymm.sum(1).view(D**2)
        phi0 = torch.randn(D**2, dtype=T.dtype, device=T.device)
        phi0 = phi0/phi0.norm()
        def Hopt(x):
            x = x.view(D, D)
            return ((Asymm.view(D*d, D) @ x).view(D, d*D) @ Asymm.permute(1,2,0).contiguous().view(d*D, D)).view(D**2)
        w = lanczos(Hopt, phi0, lanczos_steps)
    else:
        C = torch.einsum('ldr,idj->lirj', (Asymm, Asymm)).contiguous().view(D**2, D**2)
        w, _ = torch.symeig(C, eigenvectors=True) 
    lnZ2 = torch.log(w.abs().max())

    return -lnZ1 + lnZ2 

def vmps(T, d, D, Nepochs=50, Ainit=None, use_lanczos=False):

    #symmetrize
    T = (T + T.permute(3, 1, 2, 0))/2. #left-right
    T = (T + T.permute(0, 2, 1, 3))/2. #up-down
    
    if Ainit is None:
        A = torch.nn.Parameter(0.01*torch.randn(D, d, D, dtype=T.dtype, device=T.device))
    else:
        A = Ainit 

    optimizer = torch.optim.LBFGS([A], max_iter=10)
    def closure():
        optimizer.zero_grad()
        #print ('einsum', time.time()- t0)
        #print ((B-B.t()).abs().sum(), (C-C.t()).abs().sum())
        #t0 = time.time()
        loss = mpsrg(A, T.detach(), use_lanczos) # loss = -lnZ , here we optimize over A 
        #print ('mpsrg', time.time()- t0)
        #print ('            loss', loss.item())
        #t0 = time.time()
        loss.backward(retain_graph=False)
        #print ('backward', time.time()- t0)
        return loss 

    for epoch in range(Nepochs):
        loss = optimizer.step(closure)
        print ('        epoch, loss', epoch, loss.item())

    return -mpsrg(A.detach(), T, use_lanczos)  # pass lnZ out, we need to optimize over T

if __name__=='__main__':
    import argparse
    parser = argparse.ArgumentParser(description='')
    parser.add_argument("-D", type=int, default=2, help="D")
    parser.add_argument("-Dcut", type=int, default=20, help="Dcut")
    parser.add_argument("-beta", type=float, default=0.44, help="beta")
    parser.add_argument("-Nepochs", type=int, default=100, help="Nepochs")
    parser.add_argument("-float32", action='store_true', help="use float32")
    parser.add_argument("-lanczos_steps", type=int, default=0, help="lanczos steps")
    parser.add_argument("-cuda", type=int, default=-1, help="use GPU")
    args = parser.parse_args()
    device = torch.device("cpu" if args.cuda<0 else "cuda:"+str(args.cuda))
    dtype = torch.float32 if args.float32 else torch.float64
    
    #build tensor
    K = torch.tensor([args.beta], dtype=torch.float64, device=device, requires_grad=True)
    c = torch.sqrt(torch.cosh(K))
    s = torch.sqrt(torch.sinh(K))
    M = torch.stack([torch.cat([c, s]), torch.cat([c, -s])])
    T = torch.einsum('ai,aj,ak,al->ijkl', (M, M, M, M))
    
    #optimization
    lnZ = vmps(T, 2, args.Dcut, Nepochs=args.Nepochs, lanczos_steps=args.lanczos_steps)
    #recompute lnZ using optimized A
    dlnZ = torch.autograd.grad(lnZ, K, create_graph=True)[0] #  En = -d lnZ / d beta
    print (-dlnZ.item())
    #second order derivative evaluated in this way seems not correct, no effect of envioment  
    #dlnZ2 = torch.autograd.grad(dlnZ, K)[0] # Cv = beta^2 * d^2 lnZ / d beta^2 
